! start the module containing the matmul kernel
module mmul_mod

use cudafor

contains
attributes(global) subroutine mmul_global_kernel( A, B, C, N )
integer, value :: N
real :: A(N,N), B(N,N), C(N,N)
integer :: i, j, k
real :: cvalue 
! This thread computes C(i,j) = sum(A(i,:) * B(:,j))
! INSERIRE INDICI THREADS
i = 
j = 
!
if((i.le.N).and.(j.le.N)) then
!  INSERIRE CORPO COMPUTAZIONALE

endif
end subroutine mmul_global_kernel


! The host routine to drive the matrix multiplication
subroutine mmul_device( A, B, C, N )
use cudafor
integer N,ierr
real, dimension(N,N) :: A, B, C
real, device, allocatable, dimension(:,:) :: Adev,Bdev,Cdev
type(dim3) :: dimGrid, dimBlock
type(cudaEvent) :: start_event,end_event
real :: elapsed_cutime
integer grid_dimx,grid_dimy
integer mycudaerror 

allocate( Adev(N,N), Bdev(N,N), Cdev(N,N) )
Adev = A
Bdev = B

dimBlock = dim3(32, 32, 1)
! Gestione ottimizzata della griglia di thread per ogni blocco
grid_dimx = int(N/dimBlock%x)
grid_dimy = int(N/dimBlock%y)
if(mod(N,dimBlock%x) .gt. 0) grid_dimx = grid_dimx + 1
if(mod(N,dimBlock%y) .gt. 0) grid_dimy = grid_dimy + 1

dimGrid = dim3(grid_dimx,grid_dimy,1)
write(*,*) 'Gridsize :  ',grid_dimx,' x ',grid_dimy

! Creazione eventi e record di event start
! INSERIRE

! Chiamata cudaGetLastError per azzerare eventuali precedenti errori
! INSERIRE

! Chiamata kernel
call mmul_global_kernel<<<dimGrid,dimBlock>>>( Adev, Bdev, Cdev, N)

! Chiamata cudaGetLastError preceduta da sincronizzazione per rilevare eventali errori
! INSERIRE

! Record evento di stop, sincronizzazione eventi, calcolo elapsed time (e distruzione eventi)
! INSERIRE

print*,'CUDA Event elapsed time:  ',elapsed_cutime
print*,'CUDA Mflops:  ',mflops/(elapsed_cutime/1000.)

C = Cdev
deallocate( Adev, Bdev, Cdev )
end subroutine mmul_device
end module mmul_mod

subroutine mmul_host( A, B, C, N )
integer N
real, dimension(N,N) :: A, B, C
integer i,j,k
C = 0.
do j=1,N
do k=1,N
do i=1,N
   C(i,j) = C(i,j) + A(i,k)*B(k,j)
enddo
enddo
enddo
end subroutine mmul_host

program main_mmul

use mmul_mod

real, dimension(:,:),allocatable:: A,B,Ch,Cd
integer, parameter :: N = 1024
integer :: i,j
real, device, allocatable, dimension(:,:) :: Adev,Bdev,Cdev
type(dim3) :: dimGrid, dimBlock
integer cnt_err 

mflops = 2*float(N)*float(N)*float(N)/(1000.0*1000.0)

allocate(A(N,N),B(N,N),Ch(N,N),Cd(N,N))

do i=1,N
do j=1,N
   A(i,j) = (i*N+j)/N 
   B(i,j) = (i*N+j)/N 
enddo
enddo

call mmul_device( A, B, Cd, N)

call mmul_host( A, B, Ch, N)

cnt_err = 0
errore = 0.
do j=1,N
do i=1,N
   errore = max(abs(Cd(i,j)-Ch(i,j)),errore)
   if(errore.gt.1e-5) then 
      cnt_err = cnt_err + 1
      write(*,*) i,j,Cd(i,j),Ch(i,j)
      read(*,*)
   endif
enddo
enddo
if (cnt_err .eq. 0) then
   write(*,*) 'Test passed'
else
   print*,'Errori cnt_err :  ', cnt_err
endif

end
