#include <stdio.h>
#include <stdlib.h>
#include <cuda.h>
#include <cuda_runtime.h>

#define REPEAT 1

__global__ void arrayFunc(float* d_idata, float* d_jdata, float* d_odata, int size)
{
  int tid =  blockDim.x * blockIdx.x + threadIdx.x; 
  if (tid < size) {
    for(int i=0; i < REPEAT; i++)
       d_odata[tid] = d_idata[tid] * __expf(d_jdata[tid]);
  }
}

void initArrayData(float * array, float alpha, int size);
void arrayFuncCPU(const float* h_idata, const float* h_jdata, float* h_odata, int size);
void getChunkInfo(int i, int *d_offset, int *chunk_size, int *h_offset, int *chunk_stream, int nSize, int chunk_size_max, int num_chunk, int num_streams);

#define NSIZE 2097152
#define CHUNKSIZEMAX 65536
#define NUMSTREAMS 8

int 
main (void) {
  
  float *h_a, *h_b, *h_c;
  float *d_a, *d_b, *d_c;
  
  int nsize = NSIZE;
  int nThreads = 256;
  int nBlocks;

  cudaEvent_t start, end;
  float eventEtime;

  int chunk_size_max = CHUNKSIZEMAX;
  int num_streams = NUMSTREAMS;
  int num_chunk;
  int i;
  int h_offset, d_offset;
  int chunk_size, chunk_stream;

  cudaStream_t streams[NUMSTREAMS];

  // calcolo del numero di chunk
  num_chunk = (nsize-1) / chunk_size_max + 1;

  printf("Numero di elementi: %d\n", nsize);
  printf("Numero di streams:  %d\n", num_streams);
  printf("Numero di chunk:    %d\n", num_chunk);


  // allocazione e inizializzazione dei buffer di memoria su CPU
  cudaMallocHost((void**)&h_a, nsize * sizeof(float));
  cudaMallocHost((void**)&h_b, nsize * sizeof(float));
  cudaMallocHost((void**)&h_c, nsize * sizeof(float));

  initArrayData(h_a, 1.0f, nsize);
  initArrayData(h_b, 10.0f, nsize);

//-- inserire codice CUDA ----------------
  // allocazione dei buffer di memoria su GPU


  // creazione degli stream

//----------------------------------------

  // creazione dei cuda event: start, end
  cudaEventCreate(&start);  
  cudaEventCreate(&end);

  printf ("\nEseguo il calcolo su GPU ... ");

  cudaEventRecord(start,0);  

  for (i = 0; i < num_chunk; i++) {

    getChunkInfo(i, &d_offset, &chunk_size, &h_offset, &chunk_stream, nsize, chunk_size_max, num_chunk, num_streams);

//-- inserire codice CUDA ----------------
    // copia dei buffer da host a device

//----------------------------------------

    // calcolo del numero di blocchi
    nBlocks = (chunk_size-1) / nThreads + 1;

//-- inserire codice CUDA ----------------
    // lancio del kernel arrayFunc

//----------------------------------------

//-- inserire codice CUDA ----------------
    // copia del risultato da device a host

//----------------------------------------

  }

  cudaDeviceSynchronize();

  cudaEventRecord(end,0);  
  cudaEventSynchronize(end);
  cudaEventElapsedTime(&eventEtime, start, end);

  printf ("ok\n");

  printf("Tempo impiegato su GPU: %.2f ms\n", eventEtime);

  // calcolo su CPU
  printf("\nEseguo il calcolo su CPU ... ");
  float *cpuResult;
  float eventTimeCPU;
  cudaMallocHost((void**)&cpuResult, nsize * sizeof(float));
  cudaEventRecord(start,0);

  arrayFuncCPU(h_a, h_b, cpuResult, nsize);

  cudaEventRecord(end,0);
  cudaEventElapsedTime(&eventTimeCPU, start, end);
  printf ("ok\n");
  printf("Tempo impiegato su CPU: %.2f ms\n", eventTimeCPU);
  printf("\nSpeed UP CPU/GPU %.1fx\n", eventTimeCPU/eventEtime);
  
  printf("\nControllo il risultato:\n");
  printf ("h_c[0]       = %f\n", h_c[0]);
  printf ("cpuResult[0] = %f\n", cpuResult[0]);

  // liberare le risorse sul device
  for (i = 0; i< num_streams; i++)
    cudaStreamDestroy(streams[i]);
  cudaEventDestroy(start);
  cudaEventDestroy(end);
  cudaFree(d_a);
  cudaFree(d_b);
  cudaFree(d_c);

  // liberare le risorse sull'host
  cudaFreeHost(h_a);
  cudaFreeHost(h_b);
  cudaFreeHost(h_c);

  return 0;
}


void 
initArrayData(float * array, float alpha, int size)
{ 
  int i;
  for (i=0; i< size; i++) 
    array[i] = alpha * (float) rand() / (float) RAND_MAX;
    
}

void getChunkInfo(int i, int *d_offset, int *chunk_size, int *h_offset, int *chunk_stream, int nSize, int chunk_size_max, int num_chunk, int num_streams){

  int Reminder = nSize%chunk_size_max;

  *h_offset = i*chunk_size_max;
  *chunk_stream = i%num_streams;
  *chunk_size   = chunk_size_max;
  *d_offset = *chunk_stream * chunk_size_max;

  if (Reminder && (i == num_chunk-1)) *chunk_size = Reminder;

}

void arrayFuncCPU(const float* h_idata, const float* h_jdata, float* h_odata, int size)
{
  int i, j; 
  for (i=0; i<size; i++)
     for (j=0; j<REPEAT; j++)
        h_odata[i] = h_idata[i] * expf(h_jdata[i]);
}

